import numpy as np
from scipy.ndimage import zoom
from scipy.interpolate import interp1d


def reconcile(*arrays, order=0):
    """
    Make sure 1D arrays are the same length. If not, stretch them to match
    the longest.

    Args:
        arrays (ndarray): The input arrays.
        order (int): The order of the interpolation, passed to
            scipy.ndimage.zoom. Suggestion: 0 for integers and 1 for floats.

    Returns:
        tuple of ndarrays. The reconciled arrays — all of them are now the
            same length.

    Example:
        >>> a = np.array([2, 6, 7, 7, 3])
        >>> b = np.array([3, 7, 3])
        >>> reconcile(a, b, order=0)
        (array([2, 6, 7, 7, 3]), array([3, 7, 7, 3, 3]))
    """
    maxl = max(len(arr) for arr in arrays)
    out = []
    for arr in arrays:
        if len(arr) < maxl:
            out.append(zoom(arr, zoom=maxl/len(arr), order=order))
        else:
            out.append(arr)
    return tuple(out)


def interpolate(*arrays, num=50, dists=None, kind='linear'):
    """
    Linear interpolation between 1D arrays of the same length.

    Args:
        arrays (ndarray): The 1D arrays to interpolate. All must be the same
            length. You can use the `reconcile()` function to produce them.
        num (int): The number of steps to take, so will be the width (number
            of cols) of the output array.
        dists (array-like): A list or tuple or array of the distances (any
            units) between the arrays in the real world.
        kind (str): Will be passed to scipy.interpolate.interp1d, which does
            the lateral interpolation between samples.

    Returns:
        ndarray. The result, with `num` columns. The number of rows is the
            same as the number of samples in the input arrays.

    Example:
        >>> a = np.array([2, 6, 7, 7, 3])
        >>> b = np.array([3, 7, 7, 3, 3])
        >>> interp = interpolate(a, b, num=10)
        >>> interp.shape
        (5, 10)
    """
    intervals = len(arrays) - 1
    if dists is None:
        dists = intervals * [num / intervals]
    x = np.hstack([[0], np.cumsum(dists)])
    f = interp1d(x, np.stack(arrays), axis=0, kind=kind)
    return f(np.linspace(x[0], x[-1], num=num)).T


def unreconcile(arr, sizes, dists=None, order=0):
    """
    Opposite of reconcile. Restores the various profiles (the reference arrays,
    e.g. wells) to their original lengths.

    Args:
        sizes (int): The relative lengths of the profiles in the array.
            Default returns the input array.
        dists (array-like): The relative distances between the profiles in
            the array. Sum used to calculate the output width in pixels if
            the width argument is None. If not given, the distances are
            assumed to be equal.
        order (int): The order of the spline interpolation, from 0 to 3. The
            default is 0, which gives nearest neighbour interpolation. 1 gives
            linear interpolation, etc. Use 0 for ints and 1-3 for floats.

    Returns:
        ndarray. The resulting ndarray. The array contains NaNs where there
            is no interpolated data.
    """
    if np.all(sizes[0] == np.array(sizes)):
        # Nothing to do.
        return arr

    intervals = len(sizes) - 1

    if dists is None:
        eq = arr.shape[-1] // intervals
        dists = [eq] * intervals
    assert len(dists) == intervals

    maxlen = int(np.ceil(max(sizes) * arr.shape[0]))

    dist_ = np.cumsum(dists)
    idx = arr.shape[-1] * dist_ / max(dist_)
    chunks = np.split(arr, idx[:-1].astype(int), axis=-1)

    zoomed = []
    for left, right, chunk in zip(sizes[:-1], sizes[1:], chunks):
        zooms = np.linspace(left, right, chunk.shape[-1]+1)
        for z, col in zip(zooms, chunk.T):
            new_ = zoom(col, zoom=z, order=order, mode='nearest')
            pad_width = maxlen - new_.size
            new = np.pad(new_,
                         pad_width=(0, pad_width),
                         mode='constant',
                         constant_values=np.nan,
                         )
            zoomed.append(new)

    return np.array(zoomed).T


def panel(*arrays, num=50, dists=None, order=0, kind='linear'):
    """
    Interpolate an arbitrary collection of 1D arrays.

    Args:
        num (int): The number of steps to take, so will be the width (number
            of cols) of the output array.
        dists (array-like): The relative distances between the profiles in
            the array. Sum used to calculate the output width in pixels if
            the width argument is None. If not given, the distances are
            assumed to be equal.
        order (int): The order of the interpolation, passed to
            scipy.ndimage.zoom. Suggestion: 0 for integers and 1 for floats.
        kind (str): Will be passed to scipy.interpolate.interp1d, which does
            the lateral interpolation between samples.

    Returns:
        ndarray. The interpolated panel. Contains NaNs if sizes are
            non-uniform.
    """
    sizes = np.array([len(x) for x in arrays])
    sizes = sizes / np.max(sizes)
    rec = reconcile(*arrays)
    interp = interpolate(*rec, num=num, dists=dists, kind=kind)
    panel = unreconcile(interp, sizes=sizes, dists=dists, order=order)
    return panel



# -*- coding: utf-8 -*-
import unittest
import numpy as np

from bruges.reflection import reflection as avo

vp1 = 12250.
vp2 = 11600.
vs1 = 6200.
vs2 = 6650.
rho1 = 2.66
rho2 = 2.34
theta = 25

arr_vp1 = np.ones(1000)*2
arr_vp2 = np.ones(1000)*3
arr_vs1 = np.ones(1000)*1
arr_vs2 = np.ones(1000)*1.5
arr_rho1 = np.ones(1000)*2.5
arr_rho2 = np.ones(1000)*3.5
arr_theta = np.arange(40)


class AvoTest(unittest.TestCase):
    """
    Tests zoeppritz using a values from a spreadsheet, and also a
    qualitative comparison to plots made by the CREWES avo explorer
    web app. Other algorithms are then tested to be within 1% of the
    zoeppritz answer for angles < 40 degrees.
    """

    tolerance = 0.01
    reflect_rpp = avo.zoeppritz_rpp(vp1, vs1, rho1, vp2, vs2, rho2, theta)

    def test_zoeppritz(self):
        theta = 40.
        reflect = avo.zoeppritz(vp1, vs1, rho1, vp2, vs2, rho2, theta)

        # Number manually verified using
        # spreadsheet from http://tbberge.com/id63.html
        self.assertAlmostEquals(reflect, -0.112236, places=5)

    def test_zoeppritz_rpp(self):
        theta = 40.
        reflect = avo.zoeppritz(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        reflect_rpp = avo.zoeppritz_rpp(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        self.assertAlmostEquals(reflect, reflect_rpp, places=5)

    def test_akirichards(self):
        reflect = avo.akirichards(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        test = np.allclose(reflect, self.reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)

        # Test it won't complain about arrays.
        reflect = avo.akirichards(arr_vp1, arr_vs1, arr_rho1,
                                  arr_vp2, arr_vs2, arr_rho2,
                                  arr_theta)
        self.assertTrue(reflect.shape == (40, 1000))

    def test_akirichards_alt(self):
        reflect = avo.akirichards_alt(vp1, vs1, rho1, vp2,
                                      vs2, rho2, theta)
        test = np.allclose(reflect, self.reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)

        reflect = avo.akirichards_alt(arr_vp1, arr_vs1, arr_rho1,
                                      arr_vp2, arr_vs2, arr_rho2,
                                      arr_theta)
        self.assertTrue(reflect.shape == (40, 1000))

    def test_fatti(self):
        reflect = avo.fatti(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        test = np.allclose(reflect, self.reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)
        reflect = avo.fatti(arr_vp1, arr_vs1, arr_rho1,
                            arr_vp2, arr_vs2, arr_rho2, arr_theta)

    def test_shuey(self):
        reflect = avo.shuey(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        test = np.allclose(reflect, self.reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)

    def test_bortfeld(self):
        reflect = avo.bortfeld(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        test = np.allclose(reflect, self.reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)

    def test_hilterman(self):
        """
        Does not pass at theta = 25 deg.
        """
        theta = np.arange(10)
        reflect = avo.hilterman(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        reflect_rpp = avo.zoeppritz_rpp(vp1, vs1, rho1, vp2, vs2, rho2, theta)
        test = np.allclose(reflect, reflect_rpp, rtol=self.tolerance)
        self.assertTrue(test)

    def test_ca(self):
        """
        Test critical angles.
        """
        ca1, ca2 = avo.critical_angles(vp1=2300, vp2=2400, vs2=None)
        self.assertAlmostEquals(ca1, 73.40215786, places=5)
        self.assertTrue(np.isnan(ca2))

        ca1, ca2 = avo.critical_angles(vp1=1500, vp2=3200, vs2=1600)
        self.assertAlmostEquals(ca1, 27.9531869, places=5)
        self.assertAlmostEquals(ca2, 69.6358652, places=5)


if __name__ == '__main__':

    suite = unittest.TestLoader().loadTestsFromTestCase(AvoTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from .energy import energy

from .similarity import similarity

from .dipsteer import dipsteer

from .spectrogram import spectrogram

from .spectraldecomp import spectraldecomp

from .complex import instantaneous_amplitude
from .complex import reflection_strength
from .complex import envelope
from .complex import instantaneous_phase
from .complex import instantaneous_frequency
from .complex import quadrature

from .horizon import *



'''
===================
rpt.py
===================
Rock physics templates
Uses rock physics models defined in rpm.py
For Voigt-Reuss-Hill averages I define my own function here but I should
really change it and use the bounds already defined in bruges.
@author: Alessandro Amato del Monte
'''
import numpy as np
import matplotlib.pyplot as plt
from bruges.rockphysics import rpm
from bruges.rockphysics import fluidsub

def vrh(volumes,k,mu):
    '''
    Calculates Voigt-Reuss-Hill bounds.
    INPUT
    volumes: array with volumetric fractions
    k: array with bulk modulus
    mu: array with shear modulus
    OUTPUT
    k_u, k_l: upper (Voigt) and lower (Reuss) average for k
    mu_u, mu_l: upper (Voigt) and lower (Reuss) average for mu
    k0, mu0: Hill average of k and mu
    '''
    f=np.array(volumes).T
    k=np.resize(np.array(k),np.shape(f))
    mu=np.resize(np.array(mu),np.shape(f))
    ax=0 if f.ndim==1 else 1
    k_u = np.sum(f*k,axis=ax)
    k_l = 1./np.sum(f/k,axis=ax)
    mu_u = np.sum(f*mu,axis=ax)
    mu_l = 1./np.sum(f/mu,axis=ax)
    k0 = (k_u+k_l)/2.
    mu0 = (mu_u+mu_l)/2.
    return k_u, k_l, mu_u, mu_l, k0, mu0


def rpt(model='soft',vsh=0.0,fluid='gas',phic=0.4,Cn=8,P=10,f=1,cement='quartz'):
    if cement=='quartz':
        Kc, Gc = 37, 45
    elif cement=='calcite':
        Kc, Gc = 76.8, 32
    elif cement=='clay':
        Kc, Gc = 21, 7
    phi=np.linspace(0.1,phic-.1,6)
    sw=np.linspace(0,1,5)
    (Khc, Dhc) = (Kg, Dg) if fluid == 'gas' else (Ko,Do)
    K0,G0 = vrh([vsh, 1-vsh],[Ksh,Kqz],[Gsh,Gqz])[4:]
    D0 = vsh*Dsh+(1-vsh)*Dqz
    if model=='soft':
        Kdry, Gdry = rpm.softsand(K0,G0,phi,phic,Cn,P,f)
    elif model=='stiff':
         Kdry, Gdry = rpm.stiffsand(K0,G0,phi,phic,Cn,P,f)
    elif model=='cem':
         Kdry, Gdry = rpm.contactcement(K0,G0,phi,phic,Cn,Kc,Gc,scheme=2)
    elif model=='crit':
         Kdry, Gdry = rpm.critpor(K0,G0,phi,phic)

    xx=np.empty((phi.size,sw.size))
    yy=np.empty((phi.size,sw.size))

    for i,val in enumerate(sw):
        Kf = vrh([val,1-val],[Kb,Khc],[999,999])[1]
        Df = val*Db+(1-val)*Dhc
        vp,vs,rho,_= fluidsub.vels(Kdry,Gdry,K0,D0,Kf,Df,phi)
        xx[:,i]=vp*rho
        yy[:,i]=vp/vs

    plt.figure(figsize=(10,6))
    plt.plot(xx, yy, '-ok', alpha=0.3)
    plt.plot(xx.T, yy.T, '-ok', alpha=0.3)
    for i,val in enumerate(phi):
        plt.text(xx[i,-1],yy[i,-1]+.01,'$\phi={:.02f}$'.format(val), backgroundcolor='0.9')
    plt.text(xx[-1,0]-100,yy[-1,0],'$S_w={:.02f}$'.format(sw[0]),ha='right', backgroundcolor='0.9')
    plt.text(xx[-1,-1]-100,yy[-1,-1],'$S_w={:.02f}$'.format(sw[-1]),ha='right', backgroundcolor='0.9')
    plt.xlabel('Ip'), plt.ylabel('Vp/Vs')
    plt.xlim(xx.min()-xx.min()*.1,xx.max()+xx.max()*.1)
    plt.ylim(yy.min()-yy.min()*.1,yy.max()+yy.max()*.1)
    plt.title('RPT (N:G={0}, fluid={1})'.format(1-vsh, fluid))



# -*- coding: utf-8 -*-
"""
Complex trace attributes.

:copyright: 2021 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from scipy.signal import hilbert


def instantaneous_amplitude(traces):
    """
    Compute instantaneous amplitude, also known as the envelope or
    reflection strength.

    The attribute is computed over the last dimension. That is, time should
    be in the last dimension, so a 100 inline, 100 crossline seismic volume
    with 250 time slices would have shape (100, 100, 250).

    Args:
        traces (ndarray): The data array to use for calculating energy.
    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    return np.abs(hilbert(traces))

envelope = instantaneous_amplitude
reflection_strength = instantaneous_amplitude


def quadrature(traces):
    """
    Compute the quadrature trace.

    See https://wiki.seg.org/wiki/Instantaneous_attributes.

    Args:
        traces (ndarray): The data array to use for calculating energy.

    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    h = hilbert(traces)
    return np.abs(h) * np.sin(np.log(h).imag)


def instantaneous_phase(traces):
    """
    Compute the instantaneous phase of the data.

    .. math::

        \phi(t) = {\rm Im}[\ln h(t)]

    See https://wiki.seg.org/wiki/Instantaneous_attributes.

    Args:
        traces (ndarray): The data array to use for calculating energy.

    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    return np.angle(hilbert(traces))


def _inst_freq_claerbout(traces, dt):
    """
    Compute the instantaneous frequency using Claerbout's (1985) approximation.
    This is also the formulation given in Yilmaz (2001).

    Formulation from Barnes, A, 2016, Handbook of Poststack Seismic Attributes,
    SEG Books.

    Args:
        traces (ndarray): The data array to use for calculating energy.
        dt (float): The sample interval in seconds, e.g. 0.004 for 4 ms sample
            interval (250 Hz sample frequency).
    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    h = hilbert(traces)
    term = (h[1:] - h[:-1]) / (h[1:] + h[:-1])
    return (1 / (np.pi * dt)) * np.imag(term)


def _inst_freq_scheuer_oldenburg(traces, dt):
    """Instantaneous frequency after Scheuer & Oldenburg (1988).
    
    Scheuer, TE and DW Oldenburg (1988). Local phase velocity from complex seismic data.
    Geophysics 53 (12), p1503. DOI: http://dx.doi.org/10.1190/1.1442431.
    
    Formulation from Barnes, A, 2016, Handbook of Poststack Seismic Attributes,
    SEG Books:
    
    .. math::

        f_i(t) = \frac{1}{2\pi} \ \mathrm{Im} \left[\frac{h'(t)}{h(t)} \right]
               \approx \frac{1}{\pi T} \ \mathrm{Im} \left[\frac{h(t+T) - h(t)}{h(t+T) + h(t)} \right] 

    Args:
        traces (ndarray): The data array to use for calculating energy.
        dt (float): The sample interval in seconds, e.g. 0.004 for 4 ms sample
            interval (250 Hz sample frequency).
    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    y = quadrature(traces)
    expr = (traces[:-1] * y[1:] - traces[1:] * y[:-1]) / (traces[:-1] * traces[1:] + y[1:] * y[:-1])
    return (1 / (2 * np.pi * dt)) * np.arctan(expr)


def instantaneous_frequency(traces, dt, kind='so', percentile_clip=99):
    """
    Compute instantaneous frequency with a discrete approximation.

    The attribute is computed over the last dimension. That is, time should
    be in the last dimension, so a 100 inline, 100 crossline seismic volume
    with 250 time slices would have shape (100, 100, 250).

    These attributes can be noisy so a percentile clips is applied.

    Args:
        traces (ndarray): The data array to use for calculating energy.
        dt (float): The sample interval in seconds, e.g. 0.004 for 4 ms sample
            interval (250 Hz sample frequency).
        kind (str): "scipy", "claerbout" or "so" to denote a naive method from
            the SciPy docs, Claerbout's (1985) method or that of Scheuer & Oldenburg
            (1988). Claerbout's approximation is not stable above about half the
            Nyquist frequency (i.e. one quarter of the sampling frequency). The
            SciPy implementation is not recommended for seismic data.
        percentile_clip (float): Percentile at which to clip the data.
            Computed from the absolute values, clipped symmetrically
            at -p and +p, where p is the value at the 98th percentile.
    Returns:
        ndarray: An array the same dimensions as the input array.
    """
    methods = {'claerbout': _inst_freq_claerbout,
               'so': _inst_freq_scheuer_oldenburg,
               'scipy': lambda traces, dt: np.diff(instantaneous_phase(traces)) / (2.0 * np.pi * dt),
               }
    func = methods.get(kind)
    if func is None:
        m = f'{kind} is not supported, use "so" (Scheuer-Oldenburg, recommended), "claerbout" or "scipy".'
        raise NotImplementedError(m)
    f = func(traces, dt)
    p = np.percentile(np.abs(f), percentile_clip)
    return np.clip(f, a_min=-p, a_max=p)



"""
====================
rockphysicsmodels.py
====================

A bunch of rock physics models.
References are mentioned in docstrings of individual functions.
Docstrings follow numpy/scipy convention.

Alessandro Amato del Monte, March 2019
"""
import numpy as np


def critical_porosity(K0, G0, phi, phi_c=0.4):
    '''
    Critical porosity model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of dry sand for porosities below the critical porosity (phi_c).
    Above phi_c, the fluid phase is load-bearing, below phi_c the
    solid phase (mineral grains) is load-bearing.
    The equations here describe the variation of K and G
    for porosities below phi_c as straight lines.
    Critical porosity is usually 0.4 for sandstone,
    0.7 for chalk, 0.02-0.03 for granites.

    Parameters
    ----------   
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    phi_c : float, optional
        Critical porosity. Default: 0.4

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.

    References
    ----------
    Mavko et al. (2009), The Rock Physics Handbook, Cambridge University Press (p.370)
    '''
    Kdry = K0 * (1 - phi / phi_c)
    Gdry = G0 * (1 - phi / phi_c)
    return Kdry, Gdry


def hertz_mindlin(K0, G0, sigma, phi_c=0.4, Cn=8.6, f=1):
    '''
    Hertz-Mindlin model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of a dry pack of spheres subject to a hydrostatic confining pressure. 

    Parameters
    ----------   
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    phi_c : float, optional
        Critical porosity. Default: 0.4
    Cn : float, optional
        Coordination number Default: 8.6.
    f : float, optional
        Shear modulus correction factor,
        f=1 for dry pack with perfect adhesion
        between particles and f=0 for dry frictionless pack.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.

    References
    ----------
    Mavko et al. (2009), The Rock Physics Handbook, Cambridge University Press (p.246)
    '''
    sigma0 = sigma / 1e3  # converts pressure in same units as solid moduli (GPa)
    pr0 =(3*K0-2*G0) / (6*K0+2*G0)  # poisson's ratio of mineral mixture
    Khm = (sigma0*(Cn**2*(1 - phi_c)**2*G0**2) / (18
        * np.pi**2 * (1 - pr0)**2))**(1/3)
    Ghm = ((2+3*f-pr0*(1+3*f)) / (5*(2-pr0))) * ((
        sigma0 * (3 * Cn**2 * (1 - phi_c)**2 * G0**2) / (
        2 * np.pi**2 * (1 - pr0)**2)))**(1/3)
    return Khm, Ghm


def soft_sand(K0, G0, phi, sigma, phi_c=0.4, Cn=8.6, f=1):
    '''
    Soft sand, or friable sand or uncemented sand model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of poorly sorted dry sand by interpolating with the lower Hashin-Shtrikman bound
    the two end members at zero porosity and critical porosity.
    The zero porosity end member has K and G equal to mineral.
    The end member at critical porosity has K and G given by Hertz-Mindlin model.

    Parameters
    ----------   
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    phi_c : float, optional
        Critical porosity. Default: 0.4
    Cn : float, optional
        Coordination number Default: 8.6.
    f : float, optional
        Shear modulus correction factor,
        f=1 for dry pack with perfect adhesion
        between particles and f=0 for dry frictionless pack.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.

    References
    ----------
    Mavko et al. (2009), The Rock Physics Handbook, Cambridge University Press (p.258)
    '''
    Khm, Ghm = hertz_mindlin(K0, G0, sigma, phi_c, Cn, f)
    Kdry = -4/3 * Ghm + (((phi / phi_c) / (Khm + 4/3 * Ghm))
        + ((1 - phi / phi_c) / (K0 + 4/3 * Ghm)))**-1
    gxx = Ghm / 6 * ((9 * Khm + 8 * Ghm)  /  (Khm + 2 * Ghm))
    Gdry = -gxx + ((phi / phi_c) / (Ghm + gxx)
        + ((1 - phi / phi_c) / (G0 + gxx)))**-1
    return Kdry, Gdry


def stiff_sand(K0, G0, phi, sigma, phi_c=0.4, Cn=8.6, f=1):
    '''
    Stiff sand model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of stiff dry sands by interpolating with the upper Hashin-Shtrikman bound
    the two end members at zero porosity and critical porosity.
    The zero porosity end member has K and G equal to mineral.
    The end member at critical porosity has K and G given by Hertz-Mindlin model.

    Parameters
    ----------   
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    phi_c : float, optional
        Critical porosity. Default: 0.4
    Cn : float, optional
        Coordination number Default: 8.6.
    f : float, optional
        Shear modulus correction factor,
        f=1 for dry pack with perfect adhesion
        between particles and f=0 for dry frictionless pack.  
    
    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa

    References
    ----------
    Mavko et al. (2009), The Rock Physics Handbook, Cambridge University Press (p.260)
    '''
    Khm, Ghm = hertz_mindlin(K0, G0, sigma, phi_c, Cn, f)
    Kdry = -4/3 * G0 + (((phi / phi_c) / (Khm + 4/3 * G0))
        + ((1 - phi / phi_c) / (K0 + 4/3 * G0)))**-1
    tmp = G0 / 6*((9 * K0 + 8 * G0)  /  (K0 + 2 * G0))
    Gdry = -tmp + ((phi / phi_c) / (Ghm + tmp)
        + ((1 - phi / phi_c) / (G0 + tmp)))**-1
    return Kdry, Gdry


def contact_cement(K0, G0, phi, phi_c=0.4, Cn=8.6, Kc=37, Gc=45, scheme=2):
    '''
    Contact cement or cemented sand model,.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of dry sand where cement is deposited at grain contacts.
    The cement properties can be modified as well as the type of 
    cementation scheme.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    phi_c : float, optional
        Critical porosity. Default: 0.4
    Cn : float, optional
        Coordination number Default: 8.6.
    Kc, Gc : float, optional
        Cement bulk & shear modulus in GPa. Default: 37, 45.
    scheme : int, optional
        Cementation scheme, can be either 1 or 2:
        1: cement deposited at grain contacts
        2: cement as uniform layer around grains.
        Default: 2.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.

    References
    ----------
    Dvorkin-Nur (1996), Elasticity of High-Porosity Sandstones: Theory for Two North Sea Data Sets.
    Geophysics 61, no. 5 (1996).
    Mavko et al. (2009), The Rock Physics Handbook, Cambridge University Press (p.255)
    '''
    pr0 = (3 * K0 - 2 * G0) / (6 * K0 + 2 * G0)
    PRc = (3 * Kc - 2 * Gc) / (6 * Kc + 2 * Gc)
    if scheme == 1:  # scheme 1: cement deposited at grain contacts
        alpha = ((phi_c - phi) / (3 * Cn * (1 - phi_c)))**(1/4)
    else:  # scheme 2: cement evenly deposited on grain surface
        alpha = ((2 * (phi_c - phi)) / (3 * (1 - phi_c)))**(1/2)
    LambdaN = (2 * Gc * (1 - pr0) * (1 - PRc)) / (
        np.pi * G0 * (1 - 2 * PRc))
    N1 = -0.024153 * LambdaN**-1.3646
    N2 = 0.20405 * LambdaN**-0.89008
    N3 = 0.00024649 * LambdaN**-1.9864
    Sn = N1 * alpha**2 + N2 * alpha + N3
    LambdaT = Gc / (np.pi * G0)
    T1 = -10**-2 * (2.26 * pr0**2 + 2.07 * pr0
        + 2.3) * LambdaT**(0.079 * pr0**2 + 0.1754 * pr0 - 1.342)
    T2 = (0.0573 * pr0**2 + 0.0937 * pr0
        + 0.202) * LambdaT**(0.0274 * pr0**2 + 0.0529 * pr0 - 0.8765)
    T3 = 10**-4 * (9.654 * pr0**2 + 4.945 * pr0
        + 3.1) * LambdaT**(0.01867 * pr0**2 + 0.4011 * pr0 - 1.8186)
    St = T1 * alpha**2 + T2 * alpha + T3
    Kdry = 1 / 6 * Cn *(1 - phi_c ) * (Kc + 4/3 * Gc) * Sn
    Gdry = 3 / 5 * Kdry + 3 / 20 * Cn * (1 - phi_c) * Gc * St
    return Kdry, Gdry


def constant_cement(K0, G0, phi, phi_cem=0.38, phi_c=0.4, Cn=8.6, Kc=37, Gc=45, scheme=2):
    '''
    Constant cement model, Avseth et al. (2000).
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of high porosity dry sand with a certain initial cementation
    by interpolating with the lower Hashin-Shtrikman bound
    the two end members at zero porosity and critical porosity.
    The zero porosity end member has K and G equal to mineral.
    The high porosity end member has K and G given by the Contact Cement model.
    It is assumed that the porosity reduction is due to
    non-cementing material filling in the available pore space.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    phi_cem : float, optional
        Porosity at initial cementation. Default: 0.38.
    phi_c : float, optional
        Critical porosity. Default: 0.4.
    Cn : float, optional
        Coordination number. Default: 8.6.
    Kc, Gc : float, optional
        Cement bulk & shear modulus in GPa. Default: 37, 45.
    scheme : int, optional
        Cementation scheme, can be either 1 or 2:
        1: cement deposited at grain contacts
        2: cement as uniform layer around grains.
        Default: 2.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.
    
    References
    ----------
    Dvorkin et al. (2014), Seismic Reflections of Rock Properties, Cambridge University Press (p.30-31)
    '''
    # contact cement model
    Khi, Ghi = contact_cement(K0, G0, phi, phi_c=phi_c, Cn=Cn, Kc=Kc, Gc=Gc, scheme=scheme)
    # lower bound Hashin-Shtrikman starting from phi_cem
    Kcc, Gcc = contact_cement(K0, G0, phi_cem, phi_c=phi_c, Cn=Cn, Kc=Kc, Gc=Gc, scheme=scheme)
    Klo = -4/3 * Gcc + (((phi / phi_cem) / (Kcc + 4/3 * Gcc)) + (
        (1 - phi / phi_cem) / (K0 + 4/3 * Gcc)))**-1
    tmp = Gcc / 6* ((9 * Kcc + 8 * Gcc)  /  (Kcc + 2 * Gcc))
    Glo = -tmp + ((phi / phi_cem) / (Gcc + tmp) + (
        (1 - phi / phi_cem) / (G0 + tmp)))**-1
   
    # initialize empty vectors for K and G dry
    Kdry, Gdry = (np.full(phi.size, np.nan) for _ in range(2))
    # for porosities>phi_cem use [K,G]_HI = contact cement model
    # for porosities<=phi_cem use [K,G]_LO = constant cement model
    Kdry[phi > phi_cem] = Khi[phi > phi_cem]
    Kdry[phi <= phi_cem] = Klo[phi <= phi_cem]
    Gdry[phi > phi_cem] = Ghi[phi > phi_cem]
    Gdry[phi <= phi_cem] = Glo[phi <= phi_cem]
    return Kdry, Gdry


def increasing_cement(K0, G0, phi, phi_cem=0.38, phi_c=0.4, Cn=8.6, Kc=37, Gc=45, scheme=2):
    '''
    Increasing cement model (Modified Hashin-Shtrikman upper bound).
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of a dry sand with a certain initial cementation
    by interpolating with the upper Hashin-Shtrikman bound
    the two end members at zero porosity and critical porosity.
    The zero porosity end member has K and G equal to mineral.
    The high porosity end member has K and G given by the Contact Cement model.
    Probably best to avoid using if for porosities>phi_cem.
    Need to check references.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    phi_cem : float, optional
        Porosity at initial cementation. Default: 0.38.
    phi_c : float, optional
        Critical porosity. Default: 0.4.
    Cn : float, optional
        Coordination number. Default: 8.6.
    Kc, Gc : float, optional
        Cement bulk & shear modulus in GPa. Default: 37, 45.
    scheme : int, optional
        Cementation scheme, can be either 1 or 2:
        1: cement deposited at grain contacts
        2: cement as uniform layer around grains.
        Default: 2.

    Returns
    -------
    Kdry, Gdry : float or array_like
        dry rock bulk & shear modulus in GPa.
    '''
    Kcc, Gcc = contact_cement(K0, G0, phi_cem, phi_c=phi_c, Cn=Cn, Kc=Kc, Gc=Gc, scheme=scheme)
    Kdry = -4/3 * G0 + (((phi / phi_cem) / (Kcc + 4/3 * G0)) + (
        (1 - phi / phi_cem) / (K0 + 4/3 * G0)))**-1
    tmp = G0 / 6 * ((9 * K0 + 8 * G0)  /  (K0 + 2 * G0))
    Gdry = -tmp + ((phi / phi_cem) / (Gcc + tmp) + (
        (1 - phi / phi_cem) / (G0 + tmp)))**-1
    return Kdry, Gdry


def vernik_consol_sand(K0, G0, phi, sigma, b=10):
    '''
    Vernik & Kachanov Consolidated Sand Model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of consolidated dry sand subject to a hydrostatic confining pressure
    as a continuous solid containing pores and cracks.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    b : float, optional
        Slope parameter in pore shape empirical equation, range: 8-12.
        Default: 10.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.
    
    References
    ----------
    Vernik & Kachanov (2010), Modeling elastic properties of siliciclastic rocks, Geophysics v.75 n.6
    '''
    # empirical pore shape factor:
    p = 3.6 + b * phi
    q = p # true if phi>0.03
    psf = phi / (1 - phi)  # psf = pore shape factor multiplier

    # matrix properties: assuming arenites with K=35.6 GPa, G=33 GPa, Poisson's ratio nu_m = 0.146
    nu_m = 0.146
    Avm = (16 * (1 - nu_m**2) ) / ( 9 * (1 - 2 * nu_m))       # nu_m=0.146 --> Avm=2.46
    Bvm = (32 * (1 - nu_m) * (5 - nu_m)) / (45 * (2 - nu_m))  # nu_m=0.146 --> Bvm=1.59

    # crack density: inversely correlated to effective stress
    eta0 = 0.3 + 1.6 * phi  # crack density at zero stress
    d = 0.07  # compaction coefficient
    d = 0.02 + 0.003 * sigma
    cd = (eta0 * np.exp(-d * sigma)) / (1 - phi)

    # note: the presence at denominator of the factor (1 - phi) in psf and cd is needed
    # to account for the interaction effects, i.e. the presence of pores raises the average stress
    # in the matrix increasing compliance contributions of pores and cracks
    # this correction is referred to as Mori-Tanaka's scheme.
    # in this way, the original model which is a NIA (non-interaction model)
    # is extended and becomes effectively a model which does take into account interactions.
    Kdry = K0 * (1 + p * psf + Avm * cd)**-1
    Gdry = G0 * (1 + q * psf + Bvm * cd)**-1
    return Kdry, Gdry


def vernik_soft_sand_1(K0, G0, phi, sigma, phi_c=0.36, phi_con=0.26, b=10, n=2.00, m=2.05):
    '''
    Vernik & Kachanov Soft Sand Model 1.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of dry sand modeled as a granular material.
    Only applicable for porosities between the low-porosity end-member
    (at the consolidation porosity phi_con) and the high-porosity
    end-member (at the critical porosity phi_c).
    The low-porosity end member is calculated with Vernik's
    Consolidated Sand Model.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    phi_c : float, optional
        Critical porosity, range 0.30-0.42. Default: 0.36.
    phi_con : float, optional
        Consolidation porosity, range 0.22-0.30. Default: 0.26.
    b : float, optional
        Slope parameter in pore shape empirical equation, range: 8-12.
        Default: 10.
    n, m : float
        Empirical factors. Default: 2.00, 2.05.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.
    
    References
    ----------
    Vernik & Kachanov (2010), Modeling elastic properties of siliciclastic rocks, Geophysics v.75 n.6
    '''
    if isinstance(phi, np.ndarray):
        phi_edit = phi.copy()
        phi_edit[(phi_edit < phi_con) | (phi_edit > phi_c)]=np.nan
    else:
        phi_edit = np.array(phi)
        if (phi_edit < phi_con) | (phi_edit > phi_c):
            return np.nan, np.nan
    M0 = K0 + 4/3 * G0
    K_con, G_con = vernik_consol_sand(K0, G0, phi_con, sigma, b)
    M_con = K_con + 4/3 * G_con
    T = (1 - (phi_edit - phi_con) / (phi_c - phi_con))
    Mdry = M_con * T**n
    Gdry = G_con * T**m
    Kdry = Mdry - 4/3 * Gdry
    return Kdry, Gdry


def vernik_soft_sand_2(K0, G0, phi, p=20, q=20):
    '''
    Vernik & Kachanov Soft Sand Model 2.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of dry sand modeled as a granular material.
    Applicable in the entire porosity range.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    p, q : float, optional
        Pore shape factor for K and G, range: 10-45.
        Default: 20.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.
    
    References
    ----------
    Vernik & Kachanov (2010), Modeling elastic properties of siliciclastic rocks, Geophysics v.75 n.6
    '''
    M0 = K0 + 4/3 * G0
    Mdry = M0 * (1 + p * (phi / (1 - phi)))**-1
    Gdry = G0 * (1 + q * (phi / (1 - phi)))**-1
    Kdry = Mdry - 4/3 * Gdry
    return Kdry, Gdry


def vernik_sand_diagenesis(K0, G0, phi, sigma, phi_c=0.36, phi_con=0.26, b=10, n=2.00, m=2.05):
    '''
    Vernik & Kachanov Sandstone Diagenesis Model.
    This model describes the elastic behaviour (K=bulk and G=shear moduli)
    of dry sand modeled as a continuous solid containing pores and cracks
    for porosities below phi_con (consolidation porosity)
    using Vernik's Consolidated Sand Model,  and as a granular material
    for porosities above phi_con using Vernik's Soft Sand Model 1.

    Parameters
    ----------
    K0, G0 : float or array_like
        Mineral bulk & shear modulus in GPa.
    phi : float or array_like
        Porosity.
    sigma : float
        Effective stress in MPa.
    phi_c : float, optional
        Critical porosity, range 0.30-0.42. Default: 0.36.
    phi_con : float, optional
        Consolidation porosity, range 0.22-0.30. Default: 0.26.
    b : float, optional
        Slope parameter in pore shape empirical equation, range: 8-12.
        Default: 10.
    n, m : float
        Empirical factors. Default: 2.00, 2.05.

    Returns
    -------
    Kdry, Gdry : float or array_like
        Dry rock bulk & shear modulus in GPa.

    References
    ----------
    Vernik & Kachanov (2010), Modeling elastic properties of siliciclastic rocks, Geophysics v.75 n.6
    '''
    Kdry, Gdry = vernik_consol_sand(K0, G0, phi, sigma, b)
    Kdry_soft, Gdry_soft = vernik_soft_sand_1(K0, G0, phi, sigma, phi_c, phi_con, b, n, m)
    if isinstance(phi, np.ndarray):
        uu = phi>=phi_con
        Kdry[uu] = Kdry_soft[uu]
        Gdry[uu] = Gdry_soft[uu]
        return Kdry, Gdry
    else:
        if phi <= phi_con:
            return Kdry, Gdry
        else:
            return Kdry_soft, Gdry_soft


def vernik_shale(vclay, phi, rhom=2.73, rhob=1, Mqz=96, c33_clay=33.4, A=0.00284):
    '''
    Vernik & Kachanov Shale Model.
    This model describes the elastic behaviour in terms of velocities
    and density of inorganic shales.

    Parameters
    ----------
    vclay : float or array_like
        Dry clay content volume fraction.
    phi : float or array_like
        Porosity, maximum 0.40.
    rhom : float, optional
        Shale matrix density in g/cc. Default: 2.73.
    rhob : float, optional
        Brine density in g/cc. Default: 1.
    Mqz : float, optional
        P-wave elastic modulus of remaining minerals in GPa
        Default: 96.
    c33_clay : float, optional
        Anisotropic clay constant in GPa. Default: 33.4.
    A : float, optional
        Empirical coefficient for Vs. Default is good for illite/smectite/chlorite,
        can be raised up to .006 for kaolinite-rich clays.
        Default: 0.00284.
        
    Returns
    -------
    vp, vs, density : float or array_like
        P- and S-wave velocities in m/s, density in g/cc.

    Notes
    -----
    Shale matrix density (rhom) averages 2.73 +/- 0.03 g/cc at porosities below 0.25.
    It gradually varies with compaction and smectite-to-illite transition.
    A more accurate estimate can be calculated with this equation:
    rhom = 2.76+0.001*((rho-2)-230*np.exp(-4*(rho-2)))

    References
    ----------
    Vernik & Kachanov (2010), Modeling elastic properties of siliciclastic rocks, Geophysics v.75 n.6
    '''
    rho_matrix = 2.65 * (1 - vclay) + rhom * vclay
    k = 5.2 - 1.3 * vclay
    B, C = 0.287, 0.79
    c33_min = (vclay / c33_clay + (1 - vclay) / Mqz)**-1
    c33 = c33_min * (1 - phi)**k
    vp = np.sqrt(c33 / (rhom * (1 - phi) + rhob * phi))
    vs = np.sqrt(A * vp**4 + B * vp**2 - C)
    rho = rho_matrix * (1 - phi) + rhob * phi
    return vp * 1e3,vs * 1e3, rho



"""
Seismic wavelets.

:copyright: 2021 Agile Geoscience
:license: Apache 2.0
"""
import warnings
from collections import namedtuple

import numpy as np
import scipy.signal
from bruges.util import deprecated


def _get_time(duration, dt, sym=None):
    """
    Make a time vector.

    If `sym` is `True`, the time vector will have an odd number of samples,
    and will be symmetric about 0. If it's False, and the number of samples
    is even (e.g. duration = 0.016, dt = 0.004), then 0 will bot be center.
    """
    if sym is None:
        m = "In future releases, the default legacy behaviour will be removed. "
        m += "We recommend setting sym=True. This will be the default in v0.5+."
        warnings.warn(m, category=FutureWarning, stacklevel=2)
        return np.arange(-duration/2, duration/2, dt)
    
    # This business is to avoid some of the issues with `np.arange`:
    # (1) unpredictable length and (2) floating point weirdness, like
    # 1.234e-17 instead of 0. Not using `linspace` because figuring out
    # the length and offset gave me even more of a headache than this.
    n = int(duration / dt)
    odd = n % 2
    k = int(10**-np.floor(np.log10(dt)))
    dti = int(k * dt)  # integer dt
        
    if (odd and sym):
        t = np.arange(n)
    if (not odd and sym):
        t = np.arange(n + 1)
    if (odd and not sym): 
        t = np.arange(n)
    if (not odd and not sym):
        t = np.arange(n) - 1
        
    t -= t[-1] // 2
    
    return dti * t / k


def _generic(func, duration, dt, f, t=None, return_t=False, taper='blackman', sym=None):
    """
    Generic wavelet generator: applies a window to a continuous function.

    Args:
        func (function): The continuous function, taking t, f as arguments.
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Dominant frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        taper (str or function): The window or tapering function to apply.
            To use one of NumPy's functions, pass 'bartlett', 'blackman' (the
            default), 'hamming', or 'hanning'; to apply no tapering, pass
            'none'. To apply your own function, pass a function taking only
            the length of the window and returning the window function.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray. wavelet(s) with centre frequency f sampled on t. If you
            passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    f = np.asanyarray(f).reshape(-1, 1)

    # Compute time domain response.
    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed."
            warnings.warn(m, UserWarning, stacklevel=2)

    t[t == 0] = 1e-12  # Avoid division by zero.
    f[f == 0] = 1e-12  # Avoid division by zero.

    w = np.squeeze(func(t, f))

    if taper:
        tapers = {
            'bartlett': np.bartlett,
            'blackman': np.blackman,
            'hamming': np.hamming,
            'hanning': np.hanning,
            'none': lambda _: 1,
        }
        taper = tapers.get(taper, taper)
        w *= taper(t.size)

    if return_t:
        Wavelet = namedtuple('Wavelet', ['amplitude', 'time'])
        return Wavelet(w, t)
    else:
        return w


def sinc(duration, dt, f, t=None, return_t=False, taper='blackman', sym=None):
    """
    sinc function centered on t=0, with a dominant frequency of f Hz.

    If you pass a 1D array of frequencies, you get a wavelet bank in return.

    .. plot::
        plt.plot(bruges.filters.sinc(.5, 0.002, 40))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Dominant frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        taper (str or function): The window or tapering function to apply.
            To use one of NumPy's functions, pass 'bartlett', 'blackman' (the
            default), 'hamming', or 'hanning'; to apply no tapering, pass
            'none'. To apply your own function, pass a function taking only
            the length of the window and returning the window function.

    Returns:
        ndarray. sinc wavelet(s) with centre frequency f sampled on t. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    def func(t_, f_):
        return np.sin(2*np.pi*f_*t_) / (2*np.pi*f_*t_)

    return _generic(func, duration, dt, f, t, return_t, taper)


def cosine(duration, dt, f, t=None, return_t=False, taper='gaussian', sigma=None, sym=None):
    """
    With the default Gaussian window, equivalent to a 'modified Morlet'
    also sometimes called a 'Gabor' wavelet. The `bruges.filters.gabor`
    function returns a similar shape, but with a higher mean frequancy,
    somewhere between a Ricker and a cosine (pure tone).

    If you pass a 1D array of frequencies, you get a wavelet bank in return.

    .. plot::
        plt.plot(bruges.filters.cosine(.5, 0.002, 40))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Dominant frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        taper (str or function): The window or tapering function to apply.
            To use one of NumPy's functions, pass 'bartlett', 'blackman' (the
            default), 'hamming', or 'hanning'; to apply no tapering, pass
            'none'. To apply your own function, pass a function taking only
            the length of the window and returning the window function.
        sigma (float): Width of the default Gaussian window, in seconds.
            Defaults to 1/8 of the duration.

    Returns:
        ndarray. sinc wavelet(s) with centre frequency f sampled on t. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    if sigma is None:
        sigma = duration / 8

    def func(t_, f_):
        return np.cos(2 * np.pi * f_ * t_)

    def taper(length):
        return scipy.signal.gaussian(length, sigma/dt)

    return _generic(func, duration, dt, f, t, return_t, taper)


def gabor(duration, dt, f, t=None, return_t=False, sym=None):
    """
    Generates a Gabor wavelet with a peak frequency f0 at time t.

    https://en.wikipedia.org/wiki/Gabor_wavelet

    If you pass a 1D array of frequencies, you get a wavelet bank in return.

    .. plot::
        plt.plot(bruges.filters.gabor(.5, 0.002, 40))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Centre frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.

    Returns:
        ndarray. Gabor wavelet(s) with centre frequency f sampled on t. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    def func(t_, f_):
        return np.exp(-2 * f_**2 * t_**2) * np.cos(2 * np.pi * f_ * t_)

    return _generic(func, duration, dt, f, t, return_t)


def ricker(duration, dt, f, t=None, return_t=False, sym=None):
    """
    Also known as the mexican hat wavelet, models the function:

    .. math::
        A =  (1 - 2 \pi^2 f^2 t^2) e^{-\pi^2 f^2 t^2}

    If you pass a 1D array of frequencies, you get a wavelet bank in return.

    .. plot::
        plt.plot(bruges.filters.ricker(.5, 0.002, 40))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Centre frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray. Ricker wavelet(s) with centre frequency f sampled on t. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    f = np.asanyarray(f).reshape(-1, 1)

    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed."
            warnings.warn(m, UserWarning, stacklevel=2)

    pft2 = (np.pi * f * t)**2
    w = np.squeeze((1 - (2 * pft2)) * np.exp(-pft2))

    if return_t:
        RickerWavelet = namedtuple('RickerWavelet', ['amplitude', 'time'])
        return RickerWavelet(w, t)
    else:
        return w


def klauder(duration, dt, f,
            autocorrelate=True,
            t=None,
            return_t=False,
            taper='blackman',
            sym=None,
            **kwargs):
    """
    By default, gives the autocorrelation of a linear frequency modulated
    wavelet (sweep). Uses scipy.signal.chirp, adding dimensions as necessary.

    .. plot::
        plt.plot(bruges.filters.klauder(.5, 0.002, [10, 80]))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): is the sample interval in seconds (usually 0.001, 0.002,
            or 0.004)
        f (array-like): Upper and lower frequencies. Any sequence like (f1, f2).
            A list of lists will create a wavelet bank.
        autocorrelate (bool): Whether to autocorrelate the sweep(s) to create
            a wavelet. Default is `True`.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        taper (str or function): The window or tapering function to apply.
            To use one of NumPy's functions, pass 'bartlett', 'blackman' (the
            default), 'hamming', or 'hanning'; to apply no tapering, pass
            'none'. To apply your own function, pass a function taking only
            the length of the window and returning the window function.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.
        **kwargs: Further arguments are passed to scipy.signal.chirp. They are
            `method` ('linear','quadratic','logarithmic'), `phi` (phase offset
            in degrees), and `vertex_zero`.

    Returns:
        ndarray: The waveform. If you passed `return_t=True` then a tuple of
            (wavelet, t) is returned.
    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed. "
            m += "Pass None to suppress this warning."
            warnings.warn(m, UserWarning, stacklevel=2)

    t0, t1 = -duration/2, duration/2

    f = np.asanyarray(f).reshape(-1, 1)
    f1, f2 = f

    c = [scipy.signal.chirp(t, f1_+(f2_-f1_)/2., t1, f2_, **kwargs)
         for f1_, f2_
         in zip(f1, f2)]

    if autocorrelate:
        w = [np.correlate(c_, c_, mode='same') for c_ in c]

    w = np.squeeze(w) / np.amax(w)

    if taper:
        funcs = {
            'bartlett': np.bartlett,
            'blackman': np.blackman,
            'hamming': np.hamming,
            'hanning': np.hanning,
            'none': lambda x: x,
        }
        func = funcs.get(taper, taper)
        w *= func(t.size)

    if return_t:
        Sweep = namedtuple('Sweep', ['amplitude', 'time'])
        return Sweep(w, t)
    else:
        return w


sweep = klauder


def ormsby(duration, dt, f, t=None, return_t=False, sym=None):
    """
    The Ormsby wavelet requires four frequencies which together define a
    trapezoid shape in the spectrum. The Ormsby wavelet has several sidelobes,
    unlike Ricker wavelets.

    .. plot::
        plt.plot(bruges.filters.ormsby(.5, 0.002, [5, 10, 40, 80]))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (usually 0.001, 0.002,
            or 0.004).
        f (array-like): Sequence of form (f1, f2, f3, f4), or list of lists of
            frequencies, which will return a 2D wavelet bank.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray: A vector containing the Ormsby wavelet, or a bank of them. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.

    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    f = np.asanyarray(f).reshape(-1, 1)

    try:
        f1, f2, f3, f4 = f
    except ValueError:
        raise ValueError("The last dimension of the frequency array must be of size 4.")

    def numerator(f, t):
        return (np.sinc(f * t)**2) * ((np.pi * f) ** 2)

    pf43 = (np.pi * f4) - (np.pi * f3)
    pf21 = (np.pi * f2) - (np.pi * f1)

    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed."
            warnings.warn(m, UserWarning, stacklevel=2)

    w = ((numerator(f4, t)/pf43) - (numerator(f3, t)/pf43) -
         (numerator(f2, t)/pf21) + (numerator(f1, t)/pf21))

    w = np.squeeze(w) / np.amax(w)

    if return_t:
        OrmsbyWavelet = namedtuple('OrmsbyWavelet', ['amplitude', 'time'])
        return OrmsbyWavelet(w, t)
    else:
        return w


def ormsby_fft(duration, dt, f, P=(0, 0), return_t=True, sym=True):
    """
    Non-white Ormsby, with arbitary amplitudes.
    
    Can use as many points as you like. The power of f1 and f4 is assumed to be 0,
    so you only need to provide p2 and p3 (the corners). (You can actually provide
    as many f points as you like, as long as there are n - 2 matching p points.)

    .. plot::
        plt.plot(bruges.filters.ormsby(.5, 0.002, [5, 10, 40, 80]))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (usually 0.001, 0.002,
            or 0.004).
        f (array-like): Sequence of form (f1, f2, f3, f4), or list of lists of
            frequencies, which will return a 2D wavelet bank.
        P (tuple): The power of the f2 and f3 frequencies, in relative dB.
            (The magnitudes of f1 and f4 are assumed to be -∞ dB, i.e. a
            magnitude of 0.) The default power values of (0, 0) results in a
            trapezoidal spectrum and a conventional Ormsby wavelet. Pass, e.g.
            (0, -15) for a 'pink' wavelet, with more energy in the lower
            frequencies.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray: A vector containing the Ormsby wavelet, or a bank of them. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    fs = 1 / dt
    fN = fs // 2
    n = int(duration / dt)
    a = map(lambda p: 10**(p/20), P)

    # Linear interpolation of points.
    x  = np.linspace(0, int(fN), int(10*n))
    xp = [  0.] + list(f) +  [fN]
    fp = [0., 0.] + list(a) + [0., 0.]
    W = np.interp(x, xp, fp)

    # Compute inverse FFT.
    w_ = np.fft.fftshift(np.fft.irfft(W))
    L = int(w_.size // 2)
    normalize = lambda d: d / np.max(abs(d))
    w = normalize(w_[L-n//2:L+n//2+sym])
    t = _get_time(duration, dt, sym=sym)

    if return_t:
        OrmsbyWavelet = namedtuple('OrmsbyWavelet', ['amplitude', 'time'])
        return OrmsbyWavelet(w, t)
    else:
        return w


def berlage(duration, dt, f, n=2, alpha=180, phi=-np.pi/2, t=None, return_t=False, sym=None):
    """
    Generates a Berlage wavelet with a peak frequency f. Implements

    .. math::

        w(t) = AH(t) t^n \mathrm{e}^{-\alpha t} \cos(2 \pi f_0 t + \phi_0)

    as described in Aldridge, DF (1990), The Berlage wavelet, GEOPHYSICS
    55 (11), p 1508-1511. Berlage wavelets are causal, minimum phase and
    useful for modeling marine airgun sources.

    If you pass a 1D array of frequencies, you get a wavelet bank in return.

    .. plot::
        plt.plot(bruges.filters.berlage(0.5, 0.002, 40))

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (array-like): Centre frequency of the wavelet in Hz. If a sequence is
            passed, you will get a 2D array in return, one row per frequency.
        n (float): The time exponent; non-negative and real.
        alpha(float): The exponential decay factor; non-negative and real.
        phi (float): The phase.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): If True, then the function returns a tuple of
            wavelet, time-basis.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray. Berlage wavelet(s) with centre frequency f sampled on t. If
            you passed `return_t=True` then a tuple of (wavelet, t) is returned.
    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    f = np.asanyarray(f).reshape(-1, 1)
    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed."
            warnings.warn(m, UserWarning, stacklevel=2)


    H = np.heaviside(t, 0)
    w = H * t**n * np.exp(-alpha * t) * np.cos(2 * np.pi * f * t + phi)

    w = np.squeeze(w) / np.max(np.abs(w))

    if return_t:
        BerlageWavelet = namedtuple('BerlageWavelet', ['amplitude', 'time'])
        return BerlageWavelet(w, t)
    else:
        return w


def generalized(duration, dt, f, u=2, t=None, return_t=False, imag=False, sym=None):
    """
    Wang's generalized wavelet, of which the Ricker is a special case where
    u = 2. The parameter u is the order of the time-domain derivative, which
    can be a fractional derivative.

    As given by Wang (2015), Generalized seismic wavelets. GJI 203, p 1172-78.
    DOI: https://doi.org/10.1093/gji/ggv346. I am using the (more accurate)
    frequency domain method (eq 4 in that paper).

    .. plot::
        plt.plot(bruges.filters.generalized(.5, 0.002, 40, u=1.0))

    Args:
        duration (float): The length of the wavelet, in s.
        dt (float): The time sample interval in s.
        f (float or array-like): The frequency or frequencies, in Hertz.
        u (float or array-like): The fractional derivative parameter u.
        t (array-like): The time series to evaluate at, if you don't want one
            to be computed. If you pass `t` then `duration` and `dt` will be
            ignored, so we recommend passing `None` for those arguments.
        return_t (bool): Whether to return the time basis array.
        center (bool): Whether to center the wavelet on time 0.
        imag (bool): Whether to return the imaginary component as well.
        sym (bool): If True (default behaviour before v0.5) then the wavelet
            is forced to have an odd number of samples and the central sample
            is at 0 time.

    Returns:
        ndarray. If f and u are floats, the resulting wavelet has duration/dt
            = A samples. If you give f as an array of length M and u as an
            array of length N, then the resulting wavelet bank will have shape
            (M, N, A). If f or u are floats, their size will be 1, and they
            will be squeezed out: the bank is always squeezed to its minimum
            number of dimensions. If you passed `return_t=True` then a tuple
            of (wavelet, t) is returned.
    """
    if not return_t:
        m = "In future releases, return_t will be True by default."
        warnings.warn(m, FutureWarning, stacklevel=2)

    # Make sure we can do banks.
    f = np.asanyarray(f).reshape(-1, 1)
    u = np.asanyarray(u).reshape(-1, 1, 1)

    # Compute time domain response.
    if t is None:
        t = _get_time(duration, dt, sym=sym)
    else:
        if (duration is not None) or (dt is not None):
            m = "`duration` and `dt` are ignored when `t` is passed."
            warnings.warn(m, UserWarning, stacklevel=2)
        dt = t[1] - t[0]
        duration = len(t) * dt

    # Basics.
    om0 = f * 2 * np.pi
    u2 = u / 2
    df = 1 / duration
    nyquist = (1 / dt) / 2
    nf = 1 + nyquist / df
    t0 = duration / 2
    om = 2 * np.pi * np.arange(0, nyquist, df)

    # Compute the spectrum from Wang's eq 4.
    exp1 = np.exp((-om**2 / om0**2) + u2)
    exp2 = np.exp(-1j*om*t0 + 1j*np.pi * (1 + u2))
    W = (u2**(-u2)) * (om**u / om0**u) * exp1 * exp2

    w = np.fft.ifft(W, t.size)
    if not imag:
        w = w.real

    # At this point the wavelet bank has the shape (u, f, a),
    # where u is the size of u, f is the size of f, and a is
    # the number of amplitude samples we generated.
    w_max = np.max(np.abs(w), axis=-1)[:, :, None]
    w = np.squeeze(w / w_max)

    if return_t:
        GeneralizedWavelet = namedtuple('GeneralizedWavelet', ['amplitude', 'time'])
        return GeneralizedWavelet(w, t)
    else:
        return w


@deprecated('bruges.filters.wavelets.rotate_phase() is deprecated. Please use bruges.filters.rotate_phase() instead.')
def rotate_phase(w, phi, degrees=False):
    """
    Performs a phase rotation of wavelet or wavelet bank using:

    .. math::

        A = w(t)\cos(\phi) - h(t)\sin(\phi)

    where w(t) is the wavelet and h(t) is its Hilbert transform.

    The analytic signal can be written in the form S(t) = A(t)exp(j*theta(t))
    where A(t) = magnitude(hilbert(w(t))) and theta(t) = angle(hilbert(w(t))
    then a constant phase rotation phi would produce the analytic signal
    S(t) = A(t)exp(j*(theta(t) + phi)). To get the non analytic signal
    we take real(S(t)) == A(t)cos(theta(t) + phi)
    == A(t)(cos(theta(t))cos(phi) - sin(theta(t))sin(phi)) <= trig identity
    == w(t)cos(phi) - h(t)sin(phi)

    Args:
        w (ndarray): The wavelet vector, can be a 2D wavelet bank.
        phi (float): The phase rotation angle (in radians) to apply.
        degrees (bool): If phi is in degrees not radians.

    Returns:
        The phase rotated signal (or bank of signals).
    """
    if degrees:
        phi = phi * np.pi / 180.0
    a = scipy.signal.hilbert(w, axis=0)
    w = np.real(a) * np.cos(phi)  -  np.imag(a) * np.sin(phi)
    return w



# -*- coding: utf-8 -*-
import unittest
import numpy as np

from bruges.attribute import spectraldecomp
from bruges.attribute import spectrogram


class SpectraTest(unittest.TestCase):

    f1 = 100
    a1 = 1.0

    f2 = 200
    a2 = 1.5

    f3 = 450
    a3 = 2.2

    def setUp(self):

        # Make a dataset with bin centered sin waves with 3 different
        # amplitudes.
        self.duration = 10.0
        self.fs = 1024

        t = np.arange(0, self.duration, 1. / self.fs, dtype=np.double)

        w1 = 2. * np.pi * self.f1
        w2 = 2. * np.pi * self.f2
        w3 = 2. * np.pi * self.f3

        sig1 = self.a1 * np.sin(w1 * t)
        sig2 = self.a2 * np.sin(w2 * t)
        sig3 = self.a3 * np.sin(w3 * t)

        self.data = sig1 + sig2 + sig3

    def test_spectra(self):
        """
        Tests that signals show in the right bin with the
        right amplitude.
        """
        # Make the window the size of the sample rate
        window_length = self.fs

        # Test with 1.0 overlap.
        test1 = spectrogram(self.data, window_length, overlap=1.0)
        self.assertAlmostEquals(test1[0, self.f1], self.a1, places=2)
        self.assertAlmostEquals(test1[8, self.f2], self.a2, places=2)
        self.assertAlmostEquals(test1[6, self.f3], self.a3, places=2)

        # Test with 0.5 overlap.
        test2 = spectrogram(self.data, window_length, overlap=0.5)
        self.assertAlmostEquals(test2[0, self.f1], self.a1, places=2)
        self.assertAlmostEquals(test2[10, self.f2], self.a2, places=2)
        self.assertAlmostEquals(test2[15, self.f3], self.a3, places=2)

    def test_1dspectraldecomp(self):
        dt = 1.0 / self.fs
        f = (self.f1, self.f2, self.f3)
        test = spectraldecomp(self.data, window_length=1, f=f, dt=dt)
        self.assertAlmostEquals(test[5, 0] / test[5, 1], self.a1 / self.a2)
        self.assertAlmostEquals(test[5, 0] / test[5, 2], self.a1 / self.a3)

    def test_2dspectraldecomp(self):
        dt = 1.0 / self.fs
        f = (self.f1, self.f2, self.f3)
        data = np.zeros((int(self.data.size), 2))
        data[:, 0] = self.data
        data[:, 1] = self.data
        test = spectraldecomp(data, window_length=1, f=f, dt=dt)
        self.assertAlmostEquals(test[5, 0, 0] / test[5, 0, 1], self.a1/self.a2)
        self.assertAlmostEquals(test[5, 0, 0] / test[5, 0, 2], self.a1/self.a3)
        self.assertAlmostEquals(test[5, 1, 0] / test[5, 1, 1], self.a1/self.a2)
        self.assertAlmostEquals(test[5, 1, 0] / test[5, 1, 2], self.a1/self.a3)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(SpectraTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest
import numpy as np

from bruges.transform import CoordTransform


# UTM coords of 3 unique inline, crossline locations.
corners_xy = np.array([[600938.125,  6073394.5],
                       [631226.3125, 6074241.0],
                       [630720.25,   6092358.5]])

# The inline, crossline locations you just provided.
corners_ix = np.array([[99,   104],
                       [99,  1316],
                       [824, 1316]])


class CoordTransformTest(unittest.TestCase):
    """
    Tests the basic functionality of coordinate transformation.
    """

    def test_coordtransform(self):
        transform = CoordTransform(corners_ix, corners_xy)
        ix = [440, 763]
        xy = np.array([617167, 6082379])
        result = transform(ix)
        self.assertAlmostEqual(result[0], xy[0], places=-1)
        self.assertAlmostEqual(result[1], xy[1], places=-1)

        result = transform.forward(ix)
        self.assertAlmostEqual(result[0], xy[0], places=-1)
        self.assertAlmostEqual(result[1], xy[1], places=-1)

        result = transform.reverse(xy)
        self.assertAlmostEqual(result[0], ix[0], places=0)
        self.assertAlmostEqual(result[1], ix[1], places=0)


if (__name__ == '__main__'):

    suite = \
      unittest.TestLoader().loadTestsFromTestCase(CoordTransformTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest
import numpy as np

from bruges.rockphysics import elastic_impedance

vp = 2350.
vs = 1125.
rho = 2500.

theta2 = np.array([0.0, 10.0, 20.0, 30.0])

ones_vp = np.ones(1000) * vp
ones_vs = np.ones(1000) * vs
ones_rho = np.ones(1000) * rho

arr_vp = np.array([2350, 2690, 2430, 2400, 3700])
arr_vs = np.array([1125, 1300, 1240.0, 1180, 1900.0])
arr_rho = np.array([2500, 2550, 2480, 2420, 2690])

arr_vp_nan = arr_vp.astype(float)
arr_vp_nan[-2] = np.nan


class ElasticImpedanceTest(unittest.TestCase):
    """
    Tests moduli calculation against spreadsheet
    https://dl.dropboxusercontent.com/u/14965965/Elastic_moduli_formula_checking.xlsx
    """

    def test_ei_at_zero(self):
        """
        Checks that EI at zero degrees is equal to AI.
        """
        ei = elastic_impedance(vp, vs, rho, theta1=0)
        self.assertAlmostEqual(ei, vp * rho, places=-2)

    def test_ei_ninety(self):
        """
        Checks that EI at 90 degrees = (Vp/Vs)^2. This occurs when K = 0.25
        """
        ei = elastic_impedance(vp, vp/2.0, rho, theta1=90.0, use_sin=True)
        self.assertAlmostEqual(ei, (vp/(vp/2.0))**2, places=-2)

    def test_ei_log(self):
        """
        Checks that passing in 1-D arrays for vp, vs, and rho works for
        single values of theta1 as well as a 1-d vector.
        """
        x1 = elastic_impedance(ones_vp, ones_vs, ones_rho, theta1=np.array(30.0))
        x2 = elastic_impedance(ones_vp, ones_vs, ones_rho, theta1=theta2)
        x3 = elastic_impedance(arr_vp, arr_vs, arr_rho, theta1=theta2)

        self.assertEqual(x1.shape, (ones_vp.size,))
        self.assertEqual(x2.shape, (ones_vp.size, theta2.size))
        self.assertEqual(x3.shape, (arr_vp.size, theta2.size))

    def test_ei_nan(self):
        """
        Checks to see if we survuve NaNs.
        """
        ei = elastic_impedance(2350, np.nan, 2500, theta1=30.0)
        self.assertTrue(np.isnan(ei))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(ElasticImpedanceTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
